
/*

   flatland is a simple 2d physical simulator

   Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
   Written by Francois Fleuret <francois.fleuret@idiap.ch>

   This file is part of flatland

   flatland is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   flatland is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with flatland.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "canvas_cairo.h"

#include <cmath>

#define MAX(x, y) ((x >= y) ? (x) : (y))

CanvasCairo::CanvasCairo(scalar_t scale, int nb_rows, int nb_cols, CanvasCairo **ca) {
  _actual_width = 0;
  _actual_height = 0;

  for(int i = 0; i < nb_rows; i++) {
    int row_height = 0, row_width = 0;
    for(int j = 0; j < nb_cols; j++) {
      CanvasCairo *this_ca = ca[i * nb_cols + j];
      row_height = MAX(row_height, this_ca->_actual_height);
      row_width += this_ca->_actual_width;
    }
    _actual_width = MAX(_actual_width, row_width);
    _actual_height += row_height;
  }

  _data = new unsigned char [_actual_width * _actual_height * _depth];

  int x0, y0 = 0;
  for(int i = 0; i < nb_rows; i++) {
    x0 = 0;
    int row_height = 0;
    for(int j = 0; j < nb_cols; j++) {
      CanvasCairo *this_ca = ca[i * nb_cols + j];
      for(int y = 0; y < this_ca->_actual_height; y++) {
        for(int x = 0; x < this_ca->_actual_width; x++) {
          for(int d = 0; d < _depth; d++) {
            _data[(x0 + x + _actual_width * (y0 + y))* _depth + d] =
              this_ca->_data[(x + this_ca->_actual_width * y)* _depth + d];

          }
        }
      }
      row_height = MAX(row_height, this_ca->_actual_height);
      x0 += this_ca->_actual_width;
    }
    y0 += row_height;
  }

  _image = cairo_image_surface_create_for_data(_data,
                                               CAIRO_FORMAT_RGB24,
                                               _actual_width,
                                               _actual_height,
                                               _actual_width * _depth);

  _context_resource = cairo_create(_image);

  cairo_scale(_context_resource, scale, scale);

  clear();
  // cairo_set_source_rgb(_context_resource, 1.0, 1.0, 1.0);
  // cairo_set_line_width (_context_resource, 1.0);
}

CanvasCairo::CanvasCairo(scalar_t scale, int width, int height) {
  _actual_width = int(width * scale);
  _actual_height = int(height * scale);
  _scale = scale;

  _data = new unsigned char [_actual_width * _actual_height * _depth];

  _image = cairo_image_surface_create_for_data(_data,
                                               CAIRO_FORMAT_RGB24,
                                               _actual_width,
                                               _actual_height,
                                               _actual_width * _depth);

  _context_resource = cairo_create(_image);

  cairo_scale(_context_resource, scale, scale);

  cairo_set_line_width (_context_resource, 1.0);

  clear();
  // cairo_set_source_rgb(_context_resource, 1.0, 1.0, 1.0);
  // cairo_rectangle(_context_resource, 0, 0, width, height);
  // cairo_fill(_context_resource);
}

CanvasCairo::~CanvasCairo() {
  cairo_destroy(_context_resource);
  cairo_surface_destroy(_image);
  delete[] _data;
}

void CanvasCairo::clear() {
  cairo_set_source_rgb(_context_resource, 1.0, 1.0, 1.0);
  cairo_rectangle(_context_resource, 0, 0, _actual_width / _scale, _actual_height / _scale);
  cairo_fill(_context_resource);
}

void CanvasCairo::set_line_width(scalar_t w) {
  cairo_set_line_width (_context_resource, w);
}

void CanvasCairo::set_drawing_color(scalar_t r, scalar_t g, scalar_t b) {
  cairo_set_source_rgb(_context_resource, r, g, b);
}

void CanvasCairo::draw_polygon(int filled, int nb, scalar_t *x, scalar_t *y) {
  cairo_move_to(_context_resource, x[0], y[0]);
  for(int n = 0; n < nb; n++) {
    cairo_line_to(_context_resource, x[n], y[n]);
  }
  cairo_close_path(_context_resource);

  if(filled) {
    cairo_stroke_preserve(_context_resource);
    cairo_fill(_context_resource);
  } else {
    cairo_stroke(_context_resource);
  }
}

static cairo_status_t write_cairo_to_file(void *closure,
                                          const unsigned char *data,
                                          unsigned int length) {
  fwrite(data, 1, length, (FILE *) closure);
  return CAIRO_STATUS_SUCCESS;
}

void CanvasCairo::write_png(FILE *file) {
  cairo_surface_write_to_png_stream(_image, write_cairo_to_file, file);
}
